package org.processmining.plugins.log.ui.logoverview;

import java.util.ArrayList;
import java.util.List;
import java.util.SortedSet;

import javax.swing.ListModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

import org.apache.commons.collections15.list.UnmodifiableList;
import org.processmining.plugins.interactivevisualization.EID;
import org.processmining.plugins.log.ui.logoverview.filter.Filter;

public class FilterListModel implements ListModel {
	private final List<ListDataListener> listeners = new ArrayList<ListDataListener>();
	private SortedSet<EID> initial;
	private final List<Filter> filters;
	
	public FilterListModel() {
		this.initial = null;
		this.filters = new ArrayList<Filter>();
	}
	
	/**
	 * Set the initial data set to use when starting the filtering process.
	 * @param initial The set of Event IDs to use as the initial data set.
	 * In case 'null' is used, it implies that the complete log is used as the initial data set.
	 */
	public void setInitialDataSet(SortedSet<EID> events) {
		this.initial = events;
		
		//Communicate updates to listeners
		for(ListDataListener l : listeners) {
			l.contentsChanged(new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, 0, 0));
		}
	}
	
	/**
	 * Get the initial data set that is to be used as the starting data for the filtering process.
	 * @return Returns the initial data set, a sorted set of Event IDs (EID) or 'null' if the 
	 * complete log is used as the initial data set.
	 */
	public SortedSet<EID> getInitialDataSet() {
		return initial;
	}
	
	/**
	 * Add a filter to the filtering list.
	 * @param filter The filter to add to the list.
	 */
	public void addFilter(Filter filter) {
		int size = getSize(); //Current size of list. Adding a filter adds 2 to the size.
		
		filters.add(filter);
		
		ListDataEvent event = new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED, size, size+1);
		for(ListDataListener l : listeners) {
			l.contentsChanged(event);
		}
	}
	
	public void moveFilter(Filter filter, int delta) {
		
		if(filter == null) {
			throw new NullPointerException();
		}
		
		int currentIndex = filters.indexOf(filter);
		
		if(currentIndex < 0 || currentIndex + delta < 0 || currentIndex + delta >= filters.size()) {
			return; 
		}
		
		filters.remove(currentIndex);
		filters.add(currentIndex+delta, filter);
		
		ListDataEvent event = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, Math.min(currentIndex, currentIndex+delta), Math.max(currentIndex, currentIndex+delta));
		for(ListDataListener l : listeners) {
			l.contentsChanged(event);
		}
	}
	
	/**
	 * Remove a filter from the list.
	 * @param filter The filter that is to be removed.
	 */
	public void removeFilter(Filter filter) {
		
		filters.remove(filter);
		
		//Communicate "some change" to listeners. It's somewhere in the list of filters,
		// but we're not exactly sure where.
		ListDataEvent event = new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED, 1, getSize()-1);
		for(ListDataListener l : listeners) {
			l.contentsChanged(event);
		}
	}
	
	/**
	 * Remove a filter from the list.
	 * @param index Index in the list of the filter to remove.
	 */
	public void removeFilter(int index) {
		
		if(index % 2 != 1) {
			throw new IllegalArgumentException("Only filters may be removed.");
		}
		
		int filteridx = (index-1)/2;
		
		filters.remove(filteridx);
		
		//Communicate removal to listeners.
		for(ListDataListener l : listeners) {
			l.contentsChanged(new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED, index, index+1));
		}
	}
	
	/**
	 * Get all filters up to the result index 'resultidx'. (The list only contains
	 * filters that will be applied in order to acquire result 'resultidx'.)
	 * @param resultidx The index of the result in the list.
	 * @return Returns a list of Filters that are applied to acquire the result.
	 */
	public List<Filter> getFilters(int resultidx) {
		
		if(resultidx % 2 == 1) {
			//If filter provided, increase counter
			// as to point to the corresponding result.
			resultidx++;
		}
		
		int filteridx = (resultidx-2)/2;
		
		return UnmodifiableList.decorate(filters.subList(0, filteridx+1));
	}
	
	@Override
	public void addListDataListener(ListDataListener l) {
		
		if(l == null) {
			throw new NullPointerException();
		}
		
		listeners.add(l);
	}
	
	@Override
	public Object getElementAt(int index) {
		
		if(index == 0) {
			return initial;
		}
		else {
			if(index % 2 == 1) {
				int filteridx = (index-1)/2;
				
				return filters.get(filteridx);
			}
			else {
				int resultidx = (index-2)/2;
				
				return "Result of filter "+filters.get(resultidx);
			}
		}
	}
	
	@Override
	public int getSize() {
		//Size: initial data set + n*(filter + its result)
		return (1+2*filters.size());
	}
	
	@Override
	public void removeListDataListener(ListDataListener l) {
		
		if(l == null) {
			throw new NullPointerException();
		}
		
		listeners.remove(l);
	}
}
