package org.processmining.plugins.log.ui.logoverview;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.SwingConstants;

import org.apache.commons.collections15.map.UnmodifiableMap;
import org.apache.commons.collections15.set.UnmodifiableSortedSet;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.Pair;
import org.processmining.framework.util.ui.scalableview.ScalableComponent;
import org.processmining.framework.util.ui.scalableview.ScalableComponent.UpdateListener;
import org.processmining.framework.util.ui.scalableview.ScalableViewPanel;
import org.processmining.framework.util.ui.scalableview.interaction.ViewInteractionPanel;
import org.processmining.plugins.interactivevisualization.EID;
import org.processmining.plugins.interactivevisualization.InteractiveVisualization;
import org.processmining.plugins.interactivevisualization.InteractivityManager;
import org.processmining.plugins.interactivevisualization.InteractivityManagerContext;
import org.processmining.plugins.interactivevisualization.OverviewVisualizationInitializer;
import org.processmining.plugins.log.ui.logoverview.annotations.LogOverviewAddon;
import org.processmining.plugins.log.ui.logoverview.filter.Filter;
import org.processmining.plugins.log.ui.logoverview.filter.FilterLog;

/*
 * NOTE:
 * The InteractivityManagerContext does not keep track of applied filters, removed events and such.
 * By applying filters, removing events, etc. the set of event IDs used to define tags become
 * inconsistent with the events in the filtered log.
 * By recomputing the events on this level, i.e. in the LogOverviewVisualizer class - before handing
 * the set of Event IDs down to the requester, we can correct the Event IDs to the current state of
 * affairs.
 * See methods 'getAllTaggedEvents', 'getTag' and 'setTag' for the implementations.
 * 
 * NOTE:
 * Filtering is done on the original log that is stored within the LogOverviewVisualizer. So we can
 * do modifications to attributes and such and be able to keep these changes upon exporting the log.
 * However, it is not guaranteed that all plugins will be aware of these changes, i.e. changes to attributes, since they might
 * have cloned the log at an earlier stage. (Events and traces are communicated, however, their attributes are not.)
 */

public class LogOverviewVisualizer implements ScalableComponent, InteractivityManager, UpdateListener {
	
	private final List<UpdateListener> observers = new LinkedList<UpdateListener>();
	private FilterLog log;
	private UIPluginContext pluginContext;
	
	private InteractivityManagerContext managerContext;
	private InteractiveVisualization currentView = null;
	private final Map<ScalableComponent, List<Pair<ViewInteractionPanel, Integer>>> scalableViewInteractionPanels = new HashMap<ScalableComponent, List<Pair<ViewInteractionPanel, Integer>>>();
	private ScalableViewPanel scalableViewPanel = null;
	
	//Cache
	private XLogInfo summary;
	private final Map<String, Map<String, SortedSet<EID>>> cacheTags = new HashMap<String, Map<String, SortedSet<EID>>>();
	private Map<String, SortedSet<EID>> cacheVisible = null;
	private SortedSet<EID> cacheTaggedEvents = null;
	
	@Visualizer(name = "XLog Overview")
	@Plugin(name = "XLog Overview", parameterLabels = "XLog Event Log", returnTypes = JComponent.class, returnLabels = "Log Overview", userAccessible = false)
	public JComponent showLogOverview(UIPluginContext context, XLog source) {
		log = new FilterLog(source);
		summary = XLogInfoFactory.createLogInfo(log);
		managerContext = new InteractivityManagerContext(gatherExtensionTags(log, summary));
		pluginContext = context;
		
		//Gather interactive visualization initializers.
		List<OverviewVisualizationInitializer> initializerList = findAvailableInteractiveVisualizations();
		
		//Initialize visualizations.
		for(OverviewVisualizationInitializer init : initializerList) {
			
			//Initialize visualization and add the visualization to the list.
			try {
				
				//Initialize visualization.
				JComponent visualization = init.initialize(
						context.createChildContext(init.getName()),
						log,
						summary);
				
				//If visualization is ScalableViewPanel, extract real visualization component
				// Also extract interaction panels and store them in a list, such that they
				// can be added to the Log Overview ScalableViewPanel whenever the
				// visualization component is selected.
				if(visualization instanceof ScalableViewPanel) {
					ScalableViewPanel viewPanel = (ScalableViewPanel)visualization;
					
					//Get the component.
					visualization = viewPanel.getComponent();
					
					//Get the interaction panels and store them.
					List<Pair<ViewInteractionPanel, Integer>> interactionPanels = new ArrayList<Pair<ViewInteractionPanel, Integer>>();
					for(Map.Entry<ViewInteractionPanel, Integer> panel : viewPanel.getViewInteractionPanels().entrySet()) {
						
						interactionPanels.add(new Pair<ViewInteractionPanel, Integer>(panel.getKey(), panel.getValue()));
						viewPanel.removeViewInteractionPanel(panel.getKey());
					}
					scalableViewInteractionPanels.put((ScalableComponent)visualization, interactionPanels);
				}
				
				//Check if the visualization is interactive.
				if(!(visualization instanceof InteractiveVisualization)) {
					System.err.println("The visualization '"+init.getName()+"' is not an interactive visualization.");
					continue;
				}
				
				InteractiveVisualization iv = (InteractiveVisualization)visualization;
				
				iv.setManager(this);
				
				managerContext.addVisualization(init.getCategory(), init.getName(), iv);
			}
			catch(RuntimeException e) {
				e.printStackTrace();
				continue;
			}
		}
		
		//Get the available views.
		Set<InteractiveVisualization> views = managerContext.getViews();
		
		if(views.size() <= 0) {
			
			JOptionPane.showMessageDialog(null, "There are no interactive visualizations available.");
			throw new IllegalArgumentException("There are no interactive visualizations available.");
		}
		
		currentView = views.iterator().next();
		
		//Add Log Overview interaction panels.
		scalableViewPanel = new ScalableViewPanel(this);
		scalableViewPanel.addViewInteractionPanel(new LogOverviewInteractionPanel(this, managerContext.getViewNames()), SwingConstants.EAST);
		scalableViewPanel.addViewInteractionPanel(new FilterInteractionPanel(this, log), SwingConstants.EAST);
		
		//Also load interaction panels of the currently active visualization.
		// Load them only after the Log Overview interaction panels have been loaded, such that
		// the Log Overview interaction panels will always be earlier. This prevents them from
		// shifting into earlier positions when the interaction panels of the current
		// interactive visualization get removed.
		loadInteractionPanels(currentView);
		
		return scalableViewPanel;
	}
	
	/**
	 * Find all (annotated) initializers that are provided to initialize the
	 * corresponding interactive visualization.
	 * 
	 * @return Returns a list of OverviewVisualizationInitializers.
	 */
	public List<OverviewVisualizationInitializer> findAvailableInteractiveVisualizations() {
		List<OverviewVisualizationInitializer> initializers = new ArrayList<OverviewVisualizationInitializer>();
		
        for (Class<?> addonClass : pluginContext.getPluginManager().getKnownClassesAnnotatedWith(LogOverviewAddon.class)) {
        	
        	try {
        		
        		for(Class<?> interfase : addonClass.getInterfaces()) {
        			
					if(interfase.equals(OverviewVisualizationInitializer.class)) {
        				initializers.add((OverviewVisualizationInitializer)addonClass.newInstance());
        				break;
					}
        		}
        	}
        	catch (InstantiationException e) {
				e.printStackTrace();
			}
        	catch (IllegalAccessException e) {
				e.printStackTrace();
			}
        }
        
        return initializers;
	}
	
	/**
	 * Find all (annotated) filters that implement the Filter interface.
	 * 
	 * @return Returns a list of Filter instances for use in the modification framework.
	 */
	public List<Filter> findAvailableFilters() {
		List<Filter> filters = new ArrayList<Filter>();
		int errors = 0;
		
        for (Class<?> addonClass : pluginContext.getPluginManager().getKnownClassesAnnotatedWith(LogOverviewAddon.class)) {
        	
        	try {
        		
        		for(Class<?> interfase : addonClass.getInterfaces()) {
        			
					if(interfase.equals(Filter.class)) {
        				filters.add((Filter)addonClass.newInstance());
        				break;
					}
        		}
        	}
        	catch (InstantiationException e) {
        		
        		errors++;
        		e.printStackTrace();
			}
        	catch (IllegalAccessException e) {
        		
        		errors++;
				e.printStackTrace();
			}
        	catch(RuntimeException e) {
        		
        		errors++;
        		e.printStackTrace();
        	}
        }
        
        if(errors > 0) {
        	System.err.println(errors+" error(s) have occurred during the initialization of the available filters.");
        }
        
        return filters;
	}
	
	/**
	 * Get the currently active interactive visualization.
	 * 
	 * @return Returns the active interactive visualization.
	 */
	public InteractiveVisualization getCurrentView() {
		return currentView;
	}
	
	/**
	 * Change the active interactive visualization to the
	 * one specified.
	 * 
	 * @param view Interactive visualization to be active.
	 */
	public void selectView(InteractiveVisualization view) {
		
		if(view == null) {
			throw new IllegalArgumentException("'null' is not a valid view.");
		}
		
		//Unload interaction panels of previous interactive visualization.
		unloadInteractionPanels(getCurrentView());
		
		//First set the new visualization, since some interaction panels may query the scalable view panel
		// for the current scalable component and we do not want them to receive the previous visualization.
		currentView = view;
		
		//Load interaction panels for new interactive visualization.
		loadInteractionPanels(currentView);
		
		updated();
	}
	
	/**
	 * Load the interaction panels of a particular interactive visualization.
	 * This is typically the newly selected interactive visualization.
	 * Make sure interaction panels are only loaded after the currentView
	 * is set, since some interaction panels may query the ScalableViewPanel
	 * for its ScalableComponent and we don't want to accidentally return a
	 * previous component.
	 * 
	 * @param view The interactive visualization.
	 */
	private void loadInteractionPanels(InteractiveVisualization view) {
		
		//If the view is 'null' there's nothing to load.
		if(view == null) {
			return;
		}
		
		//Add the interaction panels for the newly selected view.
		if(view instanceof ScalableComponent) {
			
			List<Pair<ViewInteractionPanel, Integer>> interactionPanels = scalableViewInteractionPanels.get(view);
			if(interactionPanels != null) {
				
				for(Pair<ViewInteractionPanel, Integer> panel : interactionPanels) {
					scalableViewPanel.addViewInteractionPanel(panel.getFirst(), panel.getSecond());
				}
			}
		}
	}
	
	/**
	 * Unload the interaction panels of a particular interactive visualization.
	 * This is typically the currently active interactive visualization, just
	 * before a new interactive visualization will be selected.
	 * 
	 * @param view The interactive visualization.
	 */
	private void unloadInteractionPanels(InteractiveVisualization view) {
		
		//If view is null, there's nothing to unload.
		if(view == null) {
			return;
		}
		
		//Remove currently active interaction panels before changing view.
		if(view instanceof ScalableComponent) {
			
			List<Pair<ViewInteractionPanel, Integer>> interactionPanels = scalableViewInteractionPanels.get(view);
			if(interactionPanels != null) {
				
				for(Pair<ViewInteractionPanel, Integer> panel : interactionPanels) {
					scalableViewPanel.removeViewInteractionPanel(panel.getFirst());
				}
			}
		}
	}
	
	/**
	 * Invalidate the caches used within the LogOverviewVisualizer.
	 * Caches for:
	 *  - all requested tags.
	 *  - the visible element tag.
	 *  - the set of all enabled tags.
	 */
	private void invalidateCache() {
		cacheTags.clear();
		cacheVisible = null;
		cacheTaggedEvents = null;
	}
	
	/**
	 * Export the provided event log instance to the ProM Framework under the specified name.
	 * 
	 * @param log The event log to export.
	 * @param name The name for the newly created object.
	 */
	public void exportToFramework(XLog log, String name) {
		
		//Create a new Provided Object in the framework for the log and make it a favorite object,
		// such that the user can easily find it.
		pluginContext.getProvidedObjectManager().createProvidedObject(name, log, XLog.class, pluginContext);
		pluginContext.getGlobalContext().getResourceManager().getResourceForInstance(log).setFavorite(true);
	}
	
	/*
	 * Implementation of Interactivity Controller interface (non-Javadoc)
	 * @see org.processmining.plugins.log.ui.logoverview.InteractivityController
	 */
	
	@Override
	public void updateInteractiveViews(boolean logChanged, boolean tagsChanged) {
		
		//Invalidate caches since something has changed and the caches have become outdated.
		invalidateCache();
		
		//If the log has changed, create a new log summary for the interactive visualizations to use.
		if(logChanged) {
			summary = XLogInfoFactory.createLogInfo(log);
		}
		
		//Update all interactive visualizations.
		managerContext.updateViews(logChanged, tagsChanged);
		
		//Update all interaction panels.
		updated();
	}
	
	@Override
	public Map<String, SortedSet<EID>> getVisible() {
		Map<String, SortedSet<EID>> visible = managerContext.getVisible();
		
		if(cacheVisible == null) {
			cacheVisible = new TreeMap<String, SortedSet<EID>>();
			
			for(Map.Entry<String, SortedSet<EID>> entry : visible.entrySet()) {
				cacheVisible.put(entry.getKey(), log.convertToFiltered(entry.getValue()));
			}
		}
		
		return UnmodifiableMap.decorate(cacheVisible);
	}
	
	@Override
	public void setVisible(Map<String, SortedSet<EID>> visibleElements) {
		TreeMap<String, SortedSet<EID>> converted = new TreeMap<String, SortedSet<EID>>();
		
		for(Map.Entry<String, SortedSet<EID>> instance : visibleElements.entrySet()) {
			converted.put(instance.getKey(), log.convertToOriginal(instance.getValue()));
		}
		
		managerContext.setVisible(converted);
		
		cacheVisible = null;
	}
	
	@Override
	public Set<String> getTags() {
		return managerContext.getTags();
	}
	
	@Override
	public Set<String> getInstances(String tagName) {
		return managerContext.getInstances(tagName);
	}
	
	@Override
	public Map<String, SortedSet<EID>> getTag(String name) {
		
		if(!cacheTags.containsKey(name)) {
			Map<String, SortedSet<EID>> cache = new TreeMap<String, SortedSet<EID>>();
			
			for(Map.Entry<String, SortedSet<EID>> instance : managerContext.getTag(name).entrySet()) {
				cache.put(instance.getKey(), log.convertToFiltered(instance.getValue()));
			}
			
			cacheTags.put(name, cache);
		}

		//TODO the UnmodifiableMap only protects the outer most container from
		// changes. There is currently no protection for the sorted set of events.
		return UnmodifiableMap.decorate(cacheTags.get(name));
	}
	
	//TODO find a decent way to do this for all extensions, even non-standard ones.
	private Map<String, Map<String, SortedSet<EID>>> gatherExtensionTags(FilterLog log, XLogInfo summary) {
		XConceptExtension xconcept = XConceptExtension.instance();
		XOrganizationalExtension xorg = XOrganizationalExtension.instance();
		
		TreeMap<String, Map<String, SortedSet<EID>>> tags = new TreeMap<String, Map<String, SortedSet<EID>>>();
		
		//Using hashmaps would be faster, however we're talking very short times and by using
		// a TreeMap there is the advantage of automatic sorting by instance name.
		Map<String, SortedSet<EID>> conceptNames = new TreeMap<String, SortedSet<EID>>();
		Map<String, SortedSet<EID>> orgResources = new TreeMap<String, SortedSet<EID>>();
		for(int t = 0; t < log.size(); t++) {
			XTrace trace = log.get(t);
			
			for(int e = 0; e < trace.size(); e++) {
				XEvent event = trace.get(e);
				
				//Extract concept:name from events.
				String name = xconcept.extractName(event);
				if(name != null) {
					SortedSet<EID> set = conceptNames.get(name);
					
					if(set == null) {
						set = new TreeSet<EID>();
						conceptNames.put(name, set);
					}
					
					set.add(new EID(t, e));
				}
				
				//Extract org:resource from events.
				String resource = xorg.extractResource(event);
				if(resource != null) {
					SortedSet<EID> set = orgResources.get(resource);
					
					if(set == null) {
						set = new TreeSet<EID>();
						orgResources.put(resource, set);
					}
					
					set.add(new EID(t, e));
				}
			}
		}
		
		tags.put("concept:name", conceptNames);
		tags.put("org:resource", orgResources);
		
		return tags;
	}
	
	
	/**
	 * Additional method used to retrieve the original Event IDs.
	 * Via this method, the Event IDs will not be converted to the
	 * current state of the FilterLog.
	 * 
	 * @param name Name of the tag.
	 * @return Returns the original set of Event IDs for the given tag.
	 */
	public Map<String, SortedSet<EID>> getTagOriginals(String name) {
		return UnmodifiableMap.decorate(managerContext.getTag(name));
	}
	
	@Override
	public void setTag(String name, Map<String, SortedSet<EID>> collection) {
		Map<String, SortedSet<EID>> converted = new TreeMap<String, SortedSet<EID>>();
		
		for(Map.Entry<String, SortedSet<EID>> entry : collection.entrySet()) {
			converted.put(entry.getKey(), log.convertToOriginal(entry.getValue()));
		}
		
		managerContext.setTag(name, converted);
		
		cacheTags.remove(name); //Remove outdated cache.
	}
	
	@Override
	public void removeTag(String name) {
		
		managerContext.removeTag(name);
		
		cacheTags.remove(name); //Remove obsolete cache.
	}
	
	@Override
	public void setTagEnabled(String tagName, boolean enable) {
		
		managerContext.setTagEnabled(tagName, enable);
		
		updateInteractiveViews(false, true);
	}
	
	@Override
	public void setTagInstanceEnabled(String tagName, String instanceName, boolean enable) {
		
		managerContext.setTagInstanceEnabled(tagName, instanceName, enable);
		
		updateInteractiveViews(false, true);
	}
	
	@Override
	public boolean isTagEnabled(String tagName, String instanceName) {
		return managerContext.isTagEnabled(tagName, instanceName);
	}
	
	@Override
	public SortedSet<EID> getAllTaggedEvents() {
		
		if(cacheTaggedEvents == null) {
			cacheTaggedEvents = new TreeSet<EID>();
			
			//First, gather the enabled instances per tag.
			Map<String, List<String>> enabled = new HashMap<String, List<String>>();
			for(Pair<String, String> instance : managerContext.getEnabledTags()) {
				List<String> list = enabled.get(instance.getFirst());
				
				if(list == null) {
					list = new ArrayList<String>();
					enabled.put(instance.getFirst(), list);
				}
				
				list.add(instance.getSecond());
			}
			
			//Gather the events from enabled instances in the tags.
			for(String tagName : enabled.keySet()) {
				List<String> instances = enabled.get(tagName);
				
				Map<String, SortedSet<EID>> tag = this.getTag(tagName);
				
				for(String instance : instances) {
					cacheTaggedEvents.addAll(tag.get(instance));
				}
			}
		}
		
		return UnmodifiableSortedSet.decorate(cacheTaggedEvents);
	}
	
	@Override
	public XLogInfo getLogInfo() {
		return summary;
	}

	/*
	 * Implementation of Scalable Component interface (non-Javadoc)
	 * @see org.processmining.framework.util.ui.scalableview.ScalableComponent
	 */
	
	@Override
	public JComponent getComponent() {
		return (JComponent)currentView;
	}

	//Scaling is not implemented. If individual interactive visualizations offer scaling possibilities, we
	// can use these, since we also visualize their own interaction panels which may include zooming and
	// PIP panels. The LogOverviewVisualizer's component itself currently does not support scaling.
	@Override
	public double getScale() {
		return 1;
	}

	@Override
	public void setScale(double newScale) {}

	@Override
	public void addUpdateListener(UpdateListener listener) {
		
		if(listener == null) {
			throw new NullPointerException();
		}
		
		observers.add(listener);
	}

	@Override
	public void removeUpdateListener(UpdateListener listener) {
		
		if(listener == null) {
			return;
		}
		
		int index = observers.indexOf(listener);
		
		if(index != -1) {
			observers.remove(index);
		}
	}
	
	/*
	 * Implementation of Update Listener interface (non-Javadoc)
	 * @see org.processmining.framework.util.ui.scalableview.ScalableComponent.UpdateListener
	 */
	
	@Override
	public void updated() {
		
		for(UpdateListener observer : observers) {
			observer.updated();
		}
	}
}
