package org.processmining.plugins.log.ui.logoverview.filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.collections15.list.TreeList;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.XVisitor;
import org.processmining.plugins.interactivevisualization.EID;

/*
 * Requirement: The FilterLog is required, since the Event Identifiers
 * consist of trace index and event index. Whenever we are going to
 * delete traces or events (or add them) it will make the Event
 * Identifiers inconsistent.
 * 
 * TODO Assumptions:
 * - By cloning the log, the unique event IDs (XID) don't change.
 *   If they would, it would be impossible to relate the original
 *   with the cloned log and it would be hard to wrap the log
 *   in a FilterLog instance *and* still provide a complete XLog
 *   interface. (getting, setting, etc.)
 *   (Unless there will be some other way of checking equality.)
 * - Currently, attributes are not prevented from being changed.
 *   This means that people can change event attributes like
 *   'concept:name'. The result is that visualizations visualize
 *   information incorrectly, since it has been modified in the
 *   mean time.
 *   (although the severity is limited, since by ProM standards
 *   the plugin should clone the log before modifying it.)
 * - Setting attributes is now simply allowed.
 */

/**
 * XLog wrapper that keeps a record of removed traces. Traces are not
 * actually removed, but are inaccessible to the outside from then on.
 * 
 * This wrapper is primarily used for the purpose of filtering, while
 * keeping the original log intact for lookups of Event IDs used in
 * tagging and filtering operations.
 * 
 * NOTE: currently, list operations like retainAll and removeAll are
 * implemented in order to implement the complete XLog interface, but
 * for now are untested.
 * 
 * @author Danny van Heumen
 */
public class FilterLog implements XLog {
	
	private static List<List<Integer>> createRemainingEventsList(SortedSet<EID> events, XLog log) {
		List<List<Integer>> remaining = new ArrayList<List<Integer>>(log.size());
		
		Iterator<EID> it = events.iterator();
		int current = -1;
		List<Integer> list = null;
		while(it.hasNext()) {
			EID eid = it.next();
			
			while(current < eid.traceIdx) {	
				
				current++;
				
				//At the final iteration of the loop: create a new empty list for events to be put in.
				if(current == eid.traceIdx) {
					list = new TreeList<Integer>();
					remaining.add(list);
				}
				else {
					remaining.add(null);
				}
			}
			
			list.add(eid.eventIdx);
		}
		
		while(remaining.size() < log.size()) {
			remaining.add(null);
		}
		
		return remaining;
	}
	
	private XLog log;
	private final List<FilterTrace> traces;
	private final List<Integer> remaining = new TreeList<Integer>();
	private XEventClassifier cachedClassifier = null;
	private XLogInfo cachedInfo = null;
		
	public FilterLog(XLog log) {
		this(log, null);
	}
	
	public FilterLog(XLog log, List<List<Integer>> events) {
		this.log = log;
		
		traces = new ArrayList<FilterTrace>(log.size());
		for(int i = 0; i < this.log.size(); i++) {
			traces.add(new FilterTrace(log.get(i), (events != null ? events.get(i) : null)));
		}
		
		resetRemainingTraces();
		cachedClassifier = null;
		cachedInfo = null;
	}
	
	/**
	 * Check whether or not some log is equal to this log.
	 * A log is considered equal if both logs contain the same traces (also in the same order),
	 * and all traces are considered equal.
	 */
	public boolean equals(Object object) {
		
		if(object == this) {
			return true;
		}
		
		if(!(object instanceof XLog)) {
			return false;
		}
		
		XLog other = (XLog)object;
		
		//Check whether both logs have the same size.
		if(other.size() != this.size()) {
			return false;
		}
		
		//Check the equality of each element in the logs.
		ListIterator<XTrace> i1 = this.listIterator();
		ListIterator<XTrace> i2 = other.listIterator();
		while(i1.hasNext() && i2.hasNext()) {
		    XTrace t1 = i1.next();
		    XTrace t2 = i2.next();
		    
		    if (!(t1==null ? t2==null : t1.equals(t2))) {
		    	return false;
		    }
		}
			
		return true;
	}
	
	/**
	 * Test whether or not the log has been filtered, or whether it is still the unaltered log.
	 * @return Returns true if the log is unaltered, or false if it has been changed.
	 */
	public boolean isPristine() {
		
		//Check that the filtered log is still the same size.
		if(remaining.size() != log.size()) {
			return false;
		}
		
		//Check that the elements are still at the original location.
		for(int i = 0; i < remaining.size(); i++) {
			
			if(remaining.get(i) != i) {
				return false;
			}
		}
		
		//Check that the traces are in pristine condition.
		for(FilterTrace trace : traces) {
			
			if(!trace.isPristine()) {
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Reinsert all traces that originally existed except for traces
	 * that have a size of 0, since the XES Standard does not allow
	 * for empty traces.
	 */
	private void resetRemainingTraces() {
		
		remaining.clear();
		for(int i = 0; i < traces.size(); i++) {
			
			if(traces.get(i).size() > 0) {
				remaining.add(i);
			}
		}
	}
	
	/**
	 * Reset filter log to original log (i.e. discard earlier changes)
	 * 
	 * This will also discard changes made to individual traces as it
	 * is required that traces are not empty.
	 */
	public void reset() {
		
		//First reset all traces.
		for(FilterTrace trace : traces) {
			trace.resetRemainingEvents();
		}
		
		//Then reset list of remaining events.
		//Now nothing gets deleted since the original traces all have
		// events.
		resetRemainingTraces();
	}
	
	/**
	 * Clean up remaining traces: check all traces for empty traces and
	 * remove them. (XES standard does not allow empty traces to exist.)
	 * 
	 * @throws IllegalStateException Throws an IllegalStateException whenever the
	 * resulting event log is not valid according to the XES Standard. This may
	 * happen when an XLog instance contains 0 traces.
	 */
	public void cleanUp() {
		
		Iterator<Integer> iterator = remaining.iterator();
		while(iterator.hasNext()) {
			int i = iterator.next();
			
			if(traces.get(i).size() == 0) {
				iterator.remove();
			}
		}
		
		if(this.size() == 0) {
			throw new IllegalStateException("An XLog instance cannot have 0 traces. (XES Standard)");
		}
	}
	
	/**
	 * Clone the FilterLog.
	 * 
	 * Clones the original log, since some plugin may make (limited) changes to it.
	 * Keep the set of FilterTraces, since we want to link the logs for the purpose
	 * of manipulating the list of event indices in case of deleted events.
	 */
	public XLog clone() {
		FilterLog clone;
		
		try {
			clone = (FilterLog)super.clone();
		}
		catch (CloneNotSupportedException e) {
			
			e.printStackTrace();
			
			return null;
		}
		
		//Clone the log, because we don't want another plugin to mess up our original log.
		// Even though it is not possible to change the traces or events themselves,
		// it is still possible to change attributes, hence we don't want that to interfere
		// with the other plugins.
		clone.log = (XLog)log.clone();
		
		//Reuse traces, since we still want to set remaining events in all plugins at the same time.
		//Reuse remaining traces list, since we still want to set remaining events in all plugins at the same time.
		
		return clone;
	}
	
	/**
	 * Return the current state index of the trace of which the id is specified.
	 * @param originalTraceId Original Trace ID to search for.
	 * @return Current index of the trace or -1 if the trace is not available in the current state.
	 */
	public int indexOfOriginalId(int originalTraceId) {
		return remaining.indexOf(originalTraceId);
	}
	
	public SortedSet<EID> getRemainingEvents() {
		SortedSet<EID> remaining = new TreeSet<EID>();
		
		for(int traceIdx = 0; traceIdx < traces.size(); traceIdx++) {
			XTrace xtrace = traces.get(traceIdx);
			
			if(!(xtrace instanceof FilterTrace)) {
				throw new IllegalStateException("Encountered a trace of a type other than FilterTrace. This is not allowed.");
			}
			
			FilterTrace trace = (FilterTrace)xtrace;
			
			for(int eventIdx : trace.getRemainingEvents()) {
				remaining.add(new EID(traceIdx, eventIdx));
			}
		}
		
		return remaining;
	}
	
	/**
	 * Specify a new SortedSet of Event IDs as the set of remaining events.
	 * 
	 * @param remainingEvents The set of remaining events.
	 */
	public void setRemainingEvents(SortedSet<EID> remainingEvents) {
		List<List<Integer>> events = createRemainingEventsList(remainingEvents, log);
		
		for(int i = 0; i < traces.size(); i++) {
			traces.get(i).setRemainingEvents(events.get(i));
		}
		
		//Do clean up afterwards, since there might be empty traces.
		cleanUp();
	}

	/**
	 * Convert set of Event IDs (from the original log) to a set that corresponds
	 * to the filtered log in its current state.
	 * 
	 * Note: translating from the original set to the filtered set of Event IDs can
	 * cause problems when events have been filtered out. (The filtered set of events
	 * is a subset of the original set of events.) These events will be dropped and
	 * thus will not occur in the convert set.
	 * 
	 * @param events Set of events with Event IDs corresponding to the original log.
	 * @return Set of events with Event IDs corresponding to the filtered log.
	 */
	public SortedSet<EID> convertToFiltered(SortedSet<EID> events) {
		TreeSet<EID> filtered = new TreeSet<EID>();
		
		for(EID event : events) {
			
			int newTraceIdx = this.indexOfOriginalId(event.traceIdx);
			
			if(newTraceIdx < 0) {
				continue;
			}
			
			int newEventIdx = traces.get(event.traceIdx).indexOfOriginalId(event.eventIdx);
			
			if(newEventIdx < 0) {
				continue;
			}
			
			filtered.add(new EID(newTraceIdx, newEventIdx));
		}
		
		return filtered;
	}

	/**
	 * Convert set of Event IDs (from the filtered log in its current state) to a set
	 * that corresponds to the original log.
	 * 
	 * Note: Translating back to the original Event IDs will not cause problems,
	 * since the filtered set is a subset of the original set.
	 * 
	 * @param events Set of events with Event IDs corresponding to the filtered log.
	 * @return Set of events with Event IDs corresponding to the original log.
	 */
	public SortedSet<EID> convertToOriginal(SortedSet<EID> events) {
		TreeSet<EID> original = new TreeSet<EID>();
		
		for(EID event : events) {
			int originalTraceIdx = remaining.get(event.traceIdx);
			
			original.add(new EID(originalTraceIdx, traces.get(originalTraceIdx).getOriginalId(event.eventIdx)));
		}
				
		return original;
	}
	
	/**
	 * Generate a new (true) XLog instance using the currently active filter state.
	 * @return Returns a new instance of XLog according to the currently active filter state.
	 */
	public XLog generateXLogInstance() {
		XLog newLog = (XLog)this.log.clone();
		
		//The cloned log already possesses the correct extensions, attributes, classifiers, global event and trace attributes.
		
		//Clear the traces, since we will build up our own set of traces.
		newLog.clear();
		for(int t = 0; t < size(); t++) {
			newLog.add(traces.get(remaining.get(t)).generateXTraceInstance());
		}
		
		return newLog;
	}
	
	public boolean hasAttributes() {
		return log.hasAttributes();
	}
	
	/**
	 * Return the attributes belonging to this XLog instance.
	 */
	@Override
	public XAttributeMap getAttributes() {
		return log.getAttributes();
	}

	@Override
	public void setAttributes(XAttributeMap attributes) {
		log.setAttributes(attributes);
	}

	@Override
	public Set<XExtension> getExtensions() {
		return log.getExtensions();
	}

	@Override
	public boolean add(XTrace e) {
		throw new UnsupportedOperationException("Support for adding traces has not been implemented yet.");
	}

	@Override
	public void add(int index, XTrace element) {
		throw new UnsupportedOperationException("Support for adding traces has not been implemented yet.");
	}

	@Override
	public boolean addAll(Collection<? extends XTrace> c) {
		throw new UnsupportedOperationException("Support for adding traces has not been implemented yet.");
	}

	@Override
	public boolean addAll(int index, Collection<? extends XTrace> c) {
		throw new UnsupportedOperationException("Support for adding traces has not been implemented yet.");
	}

	@Override
	public void clear() {
		remaining.clear();
	}

	@Override
	public boolean contains(Object trace) {
		int index = traces.indexOf(trace);
		
		if(index == -1) {
			return false;
		}
		
		//Replace index in original list with index in events list.
		index = remaining.indexOf(index);
		
		return index != -1;
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		
		for(Object trace : c) {
			
			if(!contains(trace)) {
				return false;
			}
		}
		
		return true;
	}

	@Override
	public XTrace get(int index) {
		int i = remaining.get(index);
		
		if(i >= traces.size()) {
			throw new IllegalStateException("Remaining traces list is inconsistent with the traces list.");
		}
		
		return traces.get(i);
	}

	@Override
	public int indexOf(Object trace) {
		int index = traces.indexOf(trace);
		
		if(index == -1) {
			return -1;
		}
		
		//If -1 is returned at this position then the event was deleted.
		return remaining.indexOf(index);
	}

	@Override
	public boolean isEmpty() {
		return remaining.isEmpty();
	}

	@Override
	public Iterator<XTrace> iterator() {
		return new FilterLogIterator(traces, remaining);
	}

	@Override
	public int lastIndexOf(Object trace) {
		int index = traces.lastIndexOf(trace);
		
		if(index == -1) {
			return -1;
		}
		
		return remaining.lastIndexOf(index);
	}

	@Override
	public ListIterator<XTrace> listIterator() {
		return new FilterLogListIterator(this);
	}

	@Override
	public ListIterator<XTrace> listIterator(int index) {
		return new FilterLogListIterator(this, index);
	}

	@Override
	public boolean remove(Object trace) {
		int index = traces.indexOf(trace);
		
		if(index == -1) {
			return false;
		}
		
		index = remaining.indexOf(index);
		
		if(index == -1) {
			return false;
		}
		
		remaining.remove(index);
		
		return true;
	}

	@Override
	public XTrace remove(int index) {

		int i = remaining.remove(index);
		
		return traces.get(i);
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		
		for(Object trace : c) {
			remove(trace);
		}
		
		return true;
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		
		for(XTrace trace : traces) {
			
			if(!c.contains(trace)) {
				remove(trace);
			}
		}
		
		return true;
	}

	@Override
	public XTrace set(int index, XTrace trace) {
		
		if(!(trace instanceof FilterTrace)) {
			throw new IllegalArgumentException("It is not allowed to modify the traces. This is not an instance of FilterTrace and that is not allowed.");
		}
		
		int realIndex = remaining.get(index);
		
		//We need to allow 'setting' behavior to keep consistency with the FilterTrace class. As it needs setting behavior to initiate the updating of
		// attribute changes. However, since there is on Buffered XLog implementation available (yet), for now we will not fully reset the instances,
		// since this will take more processing time. Instead we take the short route of simply returning the instance that we already determined to be
		// equal.
		if(trace != null && trace.equals(traces.get(realIndex))) {
			return traces.get(realIndex);
		}
		
		throw new UnsupportedOperationException("It is only allowed to set a trace that is equal to the one already in the FilterLog instance.");
	}

	@Override
	public int size() {
		return remaining.size();
	}

	@Override
	public List<XTrace> subList(int fromIndex, int toIndex) {
		
		//Create list of indices for the requested events.
		List<Integer> indices = remaining.subList(fromIndex, toIndex);
		List<XTrace> sublist = new ArrayList<XTrace>(indices.size());
		for(int i : indices) {
			sublist.add(traces.get(i));
		}
		
		return sublist;
	}

	@Override
	public Object[] toArray() {
		XTrace[] tracearray = new XTrace[remaining.size()];
		
		for(int i = 0; i < remaining.size(); i++) {
			int traceIdx = remaining.get(i);
			
			try {
				tracearray[i] = traces.get(traceIdx);
			}
			catch(IndexOutOfBoundsException e) {
				throw new IllegalStateException("Existing/deleted events list is inconsistent with the trace list.", e);
			}
		}
		
		return tracearray;
	}

	@Override
	public <T> T[] toArray(T[] a) {
		//TODO implement this.
		throw new UnsupportedOperationException("TODO");
	}

	@Override
	public List<XEventClassifier> getClassifiers() {
		return log.getClassifiers();
	}

	@Override
	public List<XAttribute> getGlobalTraceAttributes() {
		return log.getGlobalTraceAttributes();
	}

	@Override
	public List<XAttribute> getGlobalEventAttributes() {
		return log.getGlobalEventAttributes();
	}
	
	public boolean accept(XVisitor visitor) {
		//Implementation of Visitor pattern's accept(...) method mirrored from
		//the OpenXES implementation.
		
		/*
		 * Check whether the visitor may run.
		 */
		if (visitor.precondition()) {
			
			/*
			 * Yes, it may. Now initialize.
			 */
			visitor.init(this);
			
			/*
			 * First call.
			 */
			visitor.visitLogPre(this);
			
			/*
			 * Visit the extensions.
			 */
			for (XExtension extension: this.getExtensions()) {
				extension.accept(visitor, this);
			}
			
			/*
			 * Visit the classifiers.
			 */
			for (XEventClassifier classifier: this.getClassifiers()) {
				classifier.accept(visitor, this);
			}
			
			/*
			 * Visit the attributes.
			 */
			for (XAttribute attribute: this.getAttributes().values()) {
				attribute.accept(visitor, this);
			}
			
			/*
			 * Visit the traces.
			 */
			for (XTrace trace: this) {
				trace.accept(visitor, this);
			}
			
			/*
			 * Last call.
			 */
			visitor.visitLogPost(this);
			
			return true;
		}
		
		return false;
	}
	
	private class FilterLogListIterator implements ListIterator<XTrace> {
		private static final int INIT = -1;
		
		private List<XTrace> list;
		private int next = 0;
		private int history = INIT;
		
		public FilterLogListIterator(List<XTrace> list) {
			this(list, 0);
		}
		
		public FilterLogListIterator(List<XTrace> list, int next) {
			this.list = list;
			this.next = next;
		}
		
		public boolean hasNext() {
			return next < list.size();
		}
		
		public boolean hasPrevious() {
			return (next-1) >= 0;
		}
		
		public XTrace next() {
			
			if(next == list.size()) {
				throw new NoSuchElementException();
			}
			
			//Record last position.
			history = next;
			
			//Increase position.
			next++;
			
			return list.get(history);
		}
		
		public int nextIndex() {
			return next;
		}
		
		public XTrace previous() {
			
			if(next <= 0) {
				throw new NoSuchElementException();
			}
			
			//Decrease position.
			next--;
			
			//Record last position;
			history = next;
			
			return list.get(history);
		}
		
		public int previousIndex() {
			return next-1;
		}
		
		public void add(XTrace trace) {
			
			//Add the new element to the list.
			list.add(next, trace);
			
			//Increase the next element index, as not to return the just inserted element.
			next++;
		}
		
		public void remove() {
			
			if(history == INIT) {
				throw new IllegalStateException();
			}
			
			//Remove the element from the list.
			list.remove(history);
			
			//Reset the history marker.
			history = INIT;
		}
		
		public void set(XTrace trace) {
			
			if(history == INIT) {
				throw new IllegalStateException();
			}
			
			//Set element at the position of the previous element.
			list.set(history, trace);
			
			//Reset the history marker.
			history = INIT;
		}
	}

	private class FilterLogIterator implements Iterator<XTrace> {
		private static final int INIT_VALUE = -1;
		
		private final List<? extends XTrace> traces;
		private final List<Integer> remaining;
		private int current = INIT_VALUE;
		private boolean removed = false;
		
		public FilterLogIterator(List<? extends XTrace> traces, List<Integer> remaining) {
			this.traces = traces;
			this.remaining = remaining;
		}
		
		@Override
		public boolean hasNext() {
			return (current+1 < remaining.size());
		}

		@Override
		public XTrace next() {
			
			if(current+1 >= remaining.size()) {
				//We're at the end of the list.
				throw new NoSuchElementException();
			}
			
			current++;
			
			//New iteration step, removed can be used once again.
			removed = false;
			
			return traces.get(remaining.get(current));
		}

		@Override
		public void remove() {
			
			//Don't allow removing the event before the first event.
			if(removed || current == INIT_VALUE) {
				throw new IllegalStateException((removed ? "remove() has already been called for this position." : "Start iterating before using the remove() method."));
			}
			
			//"Remove" the event.
			remaining.remove(current);
			
			//Decrease pointer since an event has been removed.
			current--;
			
			//remove() can only be called once per iteration step.
			removed = true;
		}
	}

	/**
	 * Returns the cached info if the given classifier is the cached classifier.
	 * Returns null otherwise.
	 */
	public XLogInfo getInfo(XEventClassifier classifier) {
		return classifier.equals(cachedClassifier) ? cachedInfo : null;
	}
	
	/**
	 * Sets the cached classifier and info to the given objects.
	 */
	public void setInfo(XEventClassifier classifier, XLogInfo info) {
		cachedClassifier = classifier;
		cachedInfo = info;
	}
}
