package org.processmining.plugins.log.ui.logoverview.filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;
import java.util.Set;

import org.apache.commons.collections15.list.TreeList;
import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.XVisitor;

public class FilterTrace implements XTrace {
	
	private XTrace trace;
	private final List<Integer> remaining = new TreeList<Integer>();
	
	public FilterTrace(XTrace trace) {
		this(trace, null);
	}
	
	public FilterTrace(XTrace trace, List<Integer> remainingEvents) {
		this.trace = trace;
		
		if(remainingEvents == null) {
			resetRemainingEvents();
		}
		else {
			setRemainingEvents(remainingEvents);
		}
	}
	
	/**
	 * Clone the FilterTrace.
	 * 
	 * Clones the original trace, since some plugin may make (limited) changes to it.
	 * Keep the set of event indices, since we want to link the traces for the purpose
	 * of manipulating the list of event indices in case of deleted events.
	 */
	public FilterTrace clone() {
		FilterTrace clone;
		
		try {
			clone = (FilterTrace)super.clone();
		}
		catch (CloneNotSupportedException e) {
			
			e.printStackTrace();
			
			return null;
		}
		
		clone.trace = (XTrace)trace.clone(); //Clone trace, since the clone may be partially modified.
		
		//Keep the original set of event indices. (We still need to control what events have been "deleted".)
		
		return clone;
	}
	
	/**
	 * Check whether or not some trace is equal to this trace.
	 * A trace is considered equal if both traces contain the same events (also in the same order).
	 */
	public boolean equals(Object object) {
		
		if(object == this) {
			return true;
		}
		
		if(!(object instanceof XTrace)) {
			return false;
		}
		
		XTrace other = (XTrace)object;
		
		//Check whether both traces have the same size.
		if(other.size() != this.size()) {
			return false;
		}
		
		//Check the equality of each element in the traces.
		ListIterator<XEvent> i1 = this.listIterator();
		ListIterator<XEvent> i2 = other.listIterator();
		while(i1.hasNext() && i2.hasNext()) {
		    XEvent e1 = i1.next();
		    XEvent e2 = i2.next();
		    
		    if (!(e1==null ? e2==null : e1.equals(e2))) {
		    	return false;
		    }
		}
		
		//Then the traces must be equal.
		return true;
	}
	
	/**
	 * Test whether or not the trace has been filtered, or whether it is still the unaltered trace.
	 * @return Returns true if the trace is unaltered, or false if it has been changed.
	 */
	public boolean isPristine() {
		
		//Check that the filtered trace is still the same trace.
		if(remaining.size() != trace.size()) {
			return false;
		}
		
		//Check that the elements are still at the original location.
		for(int i = 0; i < remaining.size(); i++) {
			
			if(remaining.get(i) != i) {
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Return a list of remaining events for this trace.
	 * @return Returns a list of remaining events.
	 */
	public List<Integer> getRemainingEvents() {
		return remaining;
	}
	
	/**
	 * Set a new list of remaining events. (This operation overrides earlier removes.)
	 * @param remainingEvents The events that will remain.
	 */
	public void setRemainingEvents(List<Integer> remainingEvents) {
		
		remaining.clear();
		
		if(remainingEvents != null) {
			remaining.addAll(remainingEvents);
		}
	}
	
	/**
	 * Reset to the original list of events.
	 */
	public void resetRemainingEvents() {
		
		remaining.clear();
		for(int i = 0; i < trace.size(); i++) {
			remaining.add(i);
		}
	}
	
	public int indexOfOriginalId(int originalEventIdx) {
		return remaining.indexOf(originalEventIdx);
	}
	
	public int getOriginalId(int filteredIdx) {
		return remaining.get(filteredIdx);
	}
	
	public XTrace getOriginalTrace() {
		return trace;
	}
	
	/**
	 * Generate a new (true) XTrace instance according the currently active filter state.
	 * @return Returns a new instance of XTrace.
	 */
	public XTrace generateXTraceInstance() {
		XTrace newTrace = (XTrace)trace.clone();
		
		//The cloned XTrace already has the correct attributes.
		
		//Clear the events, since we will build up our own set of events.
		newTrace.clear();
		for(int e = 0; e < size(); e++) {
			newTrace.add(trace.get(remaining.get(e)));
		}
		
		return newTrace;
	}
	
	/*
	 * Implementation of XTrace interface requirements (non-Javadoc)
	 */
	
	public boolean hasAttributes() {
		return trace.hasAttributes();
	}
	
	@Override
	public XAttributeMap getAttributes() {
		return trace.getAttributes();
	}

	@Override
	public void setAttributes(XAttributeMap attributes) {
		trace.setAttributes(attributes);
	}

	@Override
	public Set<XExtension> getExtensions() {
		return trace.getExtensions();
	}

	@Override
	public boolean add(XEvent event) {
		throw new UnsupportedOperationException("Adding new events is not implemented.)");
	}

	@Override
	public void add(int index, XEvent event) {
		throw new UnsupportedOperationException("Adding new events is not implemented.)");
	}

	@Override
	public boolean addAll(Collection<? extends XEvent> c) {
		throw new UnsupportedOperationException("Adding new events is not implemented.)");
	}

	@Override
	public boolean addAll(int index, Collection<? extends XEvent> c) {
		throw new UnsupportedOperationException("Adding new events is not implemented.)");
	}

	@Override
	public void clear() {
		remaining.clear();
	}

	@Override
	public boolean contains(Object event) {
		
		int index = trace.indexOf(event);
		
		if(index == -1) {
			return false;
		}
		
		//Replace index in original list with index in events list.
		index = remaining.indexOf(index);
		
		return index != -1;
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		
		for(Object event : c) {
			
			if(!contains(event)) {
				return false;
			}
		}
		
		return true;
	}

	@Override
	public XEvent get(int index) {
		
		int i = remaining.get(index); //Might throw ArrayIndexOutOfBounds exception if list is not as long as requested index.
		
		if(i >= trace.size()) {
			throw new IllegalStateException("Remaining events list is inconsistent with the events list.");
		}
		
		return trace.get(i);
	}

	@Override
	public int indexOf(Object event) {
		int index = trace.indexOf(event);
		
		//If -1 is returned at this point then the event has never been here in the first place.
		if(index == -1) {
			return -1;
		}
		
		//If -1 is returned at this point then the event was removed at some point.
		return remaining.indexOf(index);
	}

	@Override
	public boolean isEmpty() {
		return remaining.isEmpty();
	}

	@Override
	public Iterator<XEvent> iterator() {
		return new FilterTraceIterator(trace, remaining);
	}

	@Override
	public int lastIndexOf(Object event) {
		int index = trace.lastIndexOf(event);
		
		//If -1 is returned at this point then the event has never been here in the first place.
		if(index == -1) {
			return -1;
		}
		
		//If -1 is returned at this point then the event was removed at some point.
		return remaining.lastIndexOf(index);
	}

	@Override
	public ListIterator<XEvent> listIterator() {
		return new FilterTraceListIterator(this);
	}

	@Override
	public ListIterator<XEvent> listIterator(int index) {
		return new FilterTraceListIterator(this, index);
	}

	@Override
	public boolean remove(Object event) {
		int index = trace.indexOf(event);
		
		if(index == -1) {
			return false;
		}
		
		index = remaining.indexOf(index);
		
		if(index == -1) {
			return false;
		}
		
		remaining.remove(index);
		
		return true;
	}

	@Override
	public XEvent remove(int index) {
		
		int i = remaining.remove(index);
		
		return trace.get(i);
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		
		for(Object event : c) {
			remove(event);
		}
		
		return true;
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		
		for(XEvent event : trace) {
			
			if(!c.contains(event)) {
				remove(event);
			}
		}
		
		return true;
	}

	@Override
	public XEvent set(int index, XEvent element) {
		
		int realIndex = remaining.get(index);
		
		if(element != null && element.equals(trace.get(realIndex))) {
			//We need to allow (re)setting the same instance of an event,
			// since we are required to 'set' an event in order to trigger
			// the XFastEventList to store changes to attributes.
			return trace.set(realIndex, element);
		}
		
		throw new UnsupportedOperationException("It is only allowed to set an event that is equal to the one already in the FilterTrace instance.");
	}

	@Override
	public int size() {
		return remaining.size();
	}

	@Override
	public List<XEvent> subList(int fromIndex, int toIndex) {
		
		//Create list of indices for the requested events.
		List<Integer> indices = remaining.subList(fromIndex, toIndex);
		List<XEvent> sublist = new ArrayList<XEvent>(indices.size());
		for(int i : indices) {
			sublist.add(trace.get(i));
		}
		
		return sublist;
	}

	@Override
	public Object[] toArray() {
		XEvent[] eventarray = new XEvent[remaining.size()];
		
		for(int i = 0; i < remaining.size(); i++) {
			int traceIdx = remaining.get(i);
			
			try {
				eventarray[i] = trace.get(traceIdx);
			}
			catch(IndexOutOfBoundsException e) {
				throw new IllegalStateException("Existing/deleted events list is inconsistent with the trace list.", e);
			}
		}
		
		return eventarray;
	}

	//TODO check this code, I think it might not be completely what is requested.
	@Override
	public <T> T[] toArray(T[] a) {
		
		if(a.length > remaining.size()) {
			
			for(int i = 0; i < remaining.size(); i++) {
				a[i] = (T)trace.get(remaining.get(i));
			}
			
			return a;
		}
		else {
			return (T[])toArray();
		}
	}
	
	
	private class FilterTraceListIterator implements ListIterator<XEvent> {
		private static final int INIT = -1;
		
		private List<XEvent> list;
		private int next = 0;
		private int history = INIT;
		
		public FilterTraceListIterator(List<XEvent> list) {
			this(list, 0);
		}
		
		public FilterTraceListIterator(List<XEvent> list, int next) {
			this.list = list;
			this.next = next;
		}
		
		public boolean hasNext() {
			return next < list.size();
		}
		
		public boolean hasPrevious() {
			return (next-1) >= 0;
		}
		
		public XEvent next() {
			
			if(next == list.size()) {
				throw new NoSuchElementException();
			}
			
			//Record last position.
			history = next;
			
			//Increase position.
			next++;
			
			return list.get(history);
		}
		
		public int nextIndex() {
			return next;
		}
		
		public XEvent previous() {
			
			if(next <= 0) {
				throw new NoSuchElementException();
			}
			
			//Decrease position.
			next--;
			
			//Record last position;
			history = next;
			
			return list.get(history);
		}
		
		public int previousIndex() {
			return next-1;
		}
		
		public void add(XEvent event) {
			
			//Add the new element to the list.
			list.add(next, event);
			
			//Increase the next element index, as not to return the just inserted element.
			next++;
		}
		
		public void remove() {
			
			if(history == INIT) {
				throw new IllegalStateException();
			}
			
			//Remove the element from the list.
			list.remove(history);
			
			//Reset the history marker.
			history = INIT;
		}
		
		public void set(XEvent event) {
			
			if(history == INIT) {
				throw new IllegalStateException();
			}
			
			//Set element at the position of the previous element.
			list.set(history, event);
			
			//Reset the history marker.
			history = INIT;
		}
	}
	
	/**
	 * Iterator for iterating through the events in a FilterTrace instance.
	 * 
	 * This iterator uses the list of events indices in order to
	 * skip events that have been "deleted". And therefore can be
	 * used by other plugins that want to iterate through a modified
	 * XLog.
	 * 
	 * @author Danny van Heumen
	 */
	private class FilterTraceIterator implements Iterator<XEvent> {
		private static final int INIT_VALUE = -1;
		
		private final XTrace trace;
		private final List<Integer> remaining;
		
		private int current = INIT_VALUE;
		private boolean removed = false;
		
		public FilterTraceIterator(XTrace trace, List<Integer> remaining) {
			this.trace = trace;
			this.remaining = remaining;
		}
		
		@Override
		public boolean hasNext() {
			return (current+1 < remaining.size());
		}

		@Override
		public XEvent next() {
			
			if(current+1 >= remaining.size()) {
				//We're at the end of the list.
				throw new NoSuchElementException();
			}
			
			current++;
			
			//New iteration step, removed can be used once again.
			removed = false;
			
			return trace.get(remaining.get(current));
		}

		@Override
		public void remove() {
			
			//Don't allow removing the event before the first event.
			if(removed || current == INIT_VALUE) {
				throw new IllegalStateException((removed ? "remove() has already been called for this position." : "Start iterating before using the remove() method."));
			}
			
			//"Remove" the event.
			remaining.remove(current);
			
			//Decrease pointer since an event has been removed.
			current--;
			
			//remove() can only be called once per iteration step.
			removed = true;
		}
	}

	public int insertOrdered(XEvent event) {
		throw new UnsupportedOperationException("Adding new events is not implemented.)");
	}
	
	public void accept(XVisitor visitor, XLog log) {
		//Implementation of Visitor pattern's accept(...) method mirrored from
		//the OpenXES implementation.
		
		/*
		 * First call.
		 */
		visitor.visitTracePre(this, log);
		
		/*
		 * Visit the attributes.
		 */
		for (XAttribute attribute: this.getAttributes().values()) {
			attribute.accept(visitor, this);
		}
		
		/*
		 * Visit the events.
		 */
		for (XEvent event: this) {
			event.accept(visitor, this);
		}
		
		/*
		 * Last call.
		 */
		visitor.visitTracePost(this, log);
	}
}
